import unittest

import ssl
from http.client import HTTPConnection, HTTPSConnection

from verktyg_server.ssl import make_adhoc_ssl_context
from verktyg_server import make_socket
from verktyg_server.testing import TestServer


class TestServerTestCase(unittest.TestCase):
    def test_basic(self):
        def application(environ, start_response):
            status = '200 OK'
            headers = [('Content-type', 'text/plain; charset=utf-8')]
            start_response(status, headers)
            return [b'hello world']

        with TestServer(application) as server:
            client = HTTPConnection('localhost', server.port)

            client.request('GET', '/')
            resp = client.getresponse()
            self.assertEqual(resp.status, 200)
            self.assertEqual(resp.read(), b'hello world')

            port = server.port

        # make sure server is down after leaving context
        self.assertRaises(ConnectionRefusedError, client.request, 'GET', '/')
        make_socket('localhost', port).close()

    def test_ssl(self):
        def application(environ, start_response):
            status = '200 OK'
            headers = [('Content-type', 'text/plain; charset=utf-8')]
            start_response(status, headers)
            return [b'hello ssl']

        server_ssl_context = make_adhoc_ssl_context()
        client_ssl_context = ssl._create_stdlib_context()  # TODO

        with TestServer(application, ssl_context=server_ssl_context) as server:
            client = HTTPSConnection(
                server.host, server.port, context=client_ssl_context
            )

            client.request('GET', '/')
            resp = client.getresponse()
            self.assertEqual(resp.status, 200)



"""
    verktyg_server.testing
    ~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 by Ben Mather.
    :license:
        BSD, see LICENSE for more details.
"""
from threading import Thread

from verktyg_server import make_server, make_socket


class TestServer(object):
    def __init__(
                self, app, *, threaded=False,
                request_handler=None, ssl_context=None
            ):
        self._app = app
        self._threaded = threaded
        self._request_handler = request_handler
        self._ssl_context = ssl_context

    @property
    def protocol(self):
        return 'https' if self._ssl_context else 'http'

    @property
    def address(self):
        return '%s://%s:%s/' % (self.protocol, self.host, self.port)

    def __enter__(self):
        self.host = 'localhost'

        socket = make_socket(self.host, 0, ssl_context=self._ssl_context)
        self.port = socket.getsockname()[1]

        self._server = make_server(
            socket, self._app,
            threaded=self._threaded,
            request_handler=self._request_handler,
        )

        self._thread = Thread(target=self._server.serve_forever)
        self._thread.start()

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._server.shutdown()
        self._thread.join()



"""
    verktyg_server.ssl
    ~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 by Ben Mather.
    :license:
        BSD, see LICENSE for more details.
"""
import contextlib
from uuid import uuid4
from datetime import datetime, timedelta
import os
import tempfile
import ssl

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend


def _serialize_private_key(key):
    return key.private_bytes(
        serialization.Encoding.PEM, serialization.PrivateFormat.PKCS8,
        serialization.NoEncryption()
    )


def _deserialize_private_key(key_str):
    return serialization.load_pem_private_key(
        key_str, backend=default_backend(), password=None
    )


def _serialize_certificate(cert):
    return cert.public_bytes(
        serialization.Encoding.PEM
    )


def _deserialize_certificate(cert_str):
    return x509.load_pem_x509_certificate(
        cert_str, backend=default_backend()
    )


def _key_usage_extension(
            *, digital_signature=False, content_commitment=False,
            key_encipherment=False, data_encipherment=False,
            key_agreement=False, key_cert_sign=False, crl_sign=False,
            encipher_only=False, decipher_only=False
        ):
    return x509.KeyUsage(
        digital_signature=digital_signature,
        content_commitment=content_commitment,
        key_encipherment=key_encipherment,
        data_encipherment=data_encipherment,
        key_agreement=key_agreement,
        key_cert_sign=key_cert_sign, crl_sign=crl_sign,
        encipher_only=encipher_only, decipher_only=decipher_only
    )


def generate_adhoc_ssl_pair(*, cn=None, host=None):
    now = datetime.utcnow()
    not_valid_before = now
    not_valid_after = now + timedelta(days=16)

    if cn and host or not (cn or host):
        raise ValueError("Please specify one of common name or host")

    if host is not None:
        cn = '*.{host}/CN={host}'.format(host=host)

    name = x509.Name([
        x509.NameAttribute(
            x509.NameOID.COMMON_NAME, cn,
        ),
    ])

    key = rsa.generate_private_key(65537, 2048, backend=default_backend())

    bldr = x509.CertificateBuilder()\
        .serial_number(uuid4().int)\
        .subject_name(name)\
        .issuer_name(name)\
        .not_valid_before(not_valid_before)\
        .not_valid_after(not_valid_after)\
        .add_extension(
            _key_usage_extension(key_agreement=True), critical=True
        )\
        .public_key(key.public_key())

    cert = bldr.sign(key, SHA256(), backend=default_backend())

    return _serialize_certificate(cert), _serialize_private_key(key)


def make_adhoc_ssl_context():
    """Generates an adhoc SSL context for the development server."""
    cert, key = generate_adhoc_ssl_pair(host='example.com')

    with contextlib.ExitStack() as clean_stack:
        with contextlib.ExitStack() as close_stack:
            cert_handle, cert_file = tempfile.mkstemp(
                prefix='verktyg-adhoc-', suffix='.cert.pem'
            )
            close_stack.callback(os.close, cert_handle)
            clean_stack.callback(os.remove, cert_file)

            key_handle, key_file = tempfile.mkstemp(
                prefix='verktyg-adhoc-', suffix='.key.pem'
            )
            close_stack.callback(os.close, key_handle)
            clean_stack.callback(os.remove, key_file)

            os.write(cert_handle, cert)
            os.write(key_handle, key)

        return load_ssl_context(cert_file, key_file)


def load_ssl_context(cert_file, key_file=None):
    """Creates an SSL context from a certificate and private key file.

    :param cert_file:
        Path of the certificate to use.
    :param key_file:
        Path of the private key to use. If not given, the key will be obtained
        from the certificate file.
    """
    context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    context.load_cert_chain(cert_file, key_file)

    return context



"""
    verktyg_server
    ~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 by Ben Mather.
    :license:
        BSD, see LICENSE for more details.
"""
import sys
import urllib.parse
import ssl
import socket
from socket import getfqdn
import socketserver
from http.server import HTTPServer, BaseHTTPRequestHandler

import pkg_resources

from verktyg_server.ssl import load_ssl_context, make_adhoc_ssl_context

import logging
log = logging.getLogger('verktyg_server')


__version__ = pkg_resources.get_distribution("verktyg-server").version


class WSGIRequestHandler(BaseHTTPRequestHandler, object):
    """A request handler that implements WSGI dispatching."""

    @property
    def server_version(self):
        return 'verktyg-server/' + __version__

    def make_environ(self):
        request_url = urllib.parse.urlparse(self.path)

        path = urllib.parse.unquote_to_bytes(
            request_url.path
        ).decode('iso-8859-1')

        environ = {
            'wsgi.version':         (1, 0),
            'wsgi.input':           self.rfile,
            'wsgi.errors':          sys.stderr,
            'wsgi.multithread':     self.server.multithread,
            'wsgi.multiprocess':    self.server.multiprocess,
            'wsgi.run_once':        False,
            'verktyg.server.shutdown': self.server.shutdown,
            'SERVER_SOFTWARE':      self.server_version,
            'REQUEST_METHOD':       self.command,
            'SCRIPT_NAME':          '',
            'PATH_INFO':            path,
            'QUERY_STRING':         request_url.query,
            'CONTENT_TYPE':         self.headers.get('Content-Type', ''),
            'CONTENT_LENGTH':       self.headers.get('Content-Length', ''),
            'REMOTE_ADDR':          self.client_address[0],
            'REMOTE_PORT':          self.client_address[1],
            'SERVER_NAME':          self.server.server_address[0],
            'SERVER_PORT':          str(self.server.server_address[1]),
            'SERVER_PROTOCOL':      self.request_version
        }

        for key, value in self.headers.items():
            key = 'HTTP_' + key.upper().replace('-', '_')
            if key not in ('HTTP_CONTENT_TYPE', 'HTTP_CONTENT_LENGTH'):
                environ[key] = value

        if request_url.netloc:
            environ['HTTP_HOST'] = request_url.netloc

        if hasattr(self.request, 'getpeercert'):
            environ['REMOTE_CERT'] = self.request.getpeercert()

        return environ

    def run_wsgi(self):
        if self.headers.get('Expect', '').lower().strip() == '100-continue':
            self.wfile.write(b'HTTP/1.1 100 Continue\r\n\r\n')

        self.environ = environ = self.make_environ()
        headers_set = []
        headers_sent = []

        def write(data):
            assert headers_set, 'write() before start_response'
            if not headers_sent:
                status, response_headers = headers_sent[:] = headers_set
                try:
                    code, msg = status.split(None, 1)
                except ValueError:
                    code, msg = status, ""
                self.send_response(int(code), msg)
                header_keys = set()
                for key, value in response_headers:
                    self.send_header(key, value)
                    key = key.lower()
                    header_keys.add(key)
                if 'content-length' not in header_keys:
                    self.close_connection = True
                    self.send_header('Connection', 'close')
                if 'server' not in header_keys:
                    self.send_header('Server', self.version_string())
                if 'date' not in header_keys:
                    self.send_header('Date', self.date_time_string())
                self.end_headers()

            assert isinstance(data, bytes), 'applications must write bytes'
            self.wfile.write(data)
            self.wfile.flush()

        def start_response(status, response_headers, exc_info=None):
            if exc_info:
                try:
                    if headers_sent:
                        exc_type, exc_value, exc_traceback = exc_info
                        if exc_value.__traceback__ is not exc_traceback:
                            raise exc_value.with_traceback(exc_traceback)
                        raise exc_value
                finally:
                    exc_info = None
            elif headers_set:
                raise AssertionError('Headers already set')
            headers_set[:] = [status, response_headers]
            return write

        def execute(app):
            application_iter = app(environ, start_response)
            try:
                for data in application_iter:
                    write(data)
                if not headers_sent:
                    write(b'')
            finally:
                if hasattr(application_iter, 'close'):
                    application_iter.close()
                application_iter = None

        try:
            execute(self.server.app)
        except (socket.error, socket.timeout) as e:
            self.connection_dropped(e, environ)
        except Exception as e:
            if self.server.passthrough_errors:
                raise
            self.logger.error("Error on request", exc_info=True)
            try:
                # if we haven't yet sent the headers but they are set
                # we roll back to be able to set them again.
                if not headers_sent:
                    del headers_set[:]
                execute(self.render_error)
            except Exception:
                self.logger.exception("error recovering from failed response")

    def render_error(self, environ, start_response):
        status = '500 Internal Server Error'
        headers = [('Content-type', 'text/html')]
        start_response(status, headers)
        return [b"<h1>Internal Server Error</h1>"]

    def handle(self):
        """Handles a request ignoring dropped connections."""
        rv = None
        try:
            rv = BaseHTTPRequestHandler.handle(self)
        except (socket.error, socket.timeout, ssl.SSLError) as e:
            self.connection_dropped(e)
        return rv

    def connection_dropped(self, error, environ=None):
        """Called if the connection was closed by the client.  By default
        nothing happens.
        """

    def handle_one_request(self):
        """Handle a single HTTP request."""
        self.raw_requestline = self.rfile.readline()
        if not self.raw_requestline:
            self.close_connection = 1
        elif self.parse_request():
            return self.run_wsgi()

    def send_response(self, code, message=None):
        """Send the response header and log the response code."""
        self.log_request(code)
        if message is None:
            message = code in self.responses and self.responses[code][0] or ''
        if self.request_version != 'HTTP/0.9':
            hdr = "%s %d %s\r\n" % (self.protocol_version, code, message)
            self.wfile.write(hdr.encode('ascii'))

    def version_string(self):
        return BaseHTTPRequestHandler.version_string(self).strip()

    def address_string(self):
        return self.environ['REMOTE_ADDR']

    def log_request(self, code='-', size='-'):
        self.logger.info('%r %s %s', self.requestline, code, size)

    def log_error(self, *args):
        self.logger.error(format, *args)

    def log_message(self, format, *args):
        self.logger.info(format, *args)

    @property
    def logger(self):
        return self.server.logger


class BaseWSGIServer(HTTPServer, object):
    """Simple single-threaded, single-process WSGI server."""
    multithread = False
    multiprocess = False
    request_queue_size = 128

    def __init__(
                self, socket, app, *, handler=None,
                passthrough_errors=False, logger=None
            ):
        if logger is None:
            logger = 'verktyg-server'
        if isinstance(logger, str):
            logger = logging.getLogger(logger)
        self.logger = logger

        if handler is None:
            handler = WSGIRequestHandler

        self.socket = socket
        server_address = self.socket.getsockname()
        socketserver.BaseServer.__init__(self, server_address, handler)

        host, port = server_address[:2]
        self.server_name = getfqdn(host)
        self.server_port = port

        self.app = app
        self.passthrough_errors = passthrough_errors

    def log(self, type, message, *args):
        log.log(type, message, *args)

    def serve_forever(self):
        try:
            HTTPServer.serve_forever(self)
        except KeyboardInterrupt:
            pass
        finally:
            self.server_close()

    def handle_error(self, request, client_address):
        if self.passthrough_errors:
            raise
        else:
            return HTTPServer.handle_error(self, request, client_address)

    def get_request(self):
        con, info = self.socket.accept()
        return con, info


class ThreadedWSGIServer(socketserver.ThreadingMixIn, BaseWSGIServer):
    """A WSGI server that does threading."""
    multithread = True


class ForkingWSGIServer(socketserver.ForkingMixIn, BaseWSGIServer):
    """A WSGI server that does forking."""
    multiprocess = True

    def __init__(
                self, socket, app, *, processes=40, handler=None,
                passthrough_errors=False
            ):
        BaseWSGIServer.__init__(
            self, socket, app, handler=handler,
            passthrough_errors=passthrough_errors
        )
        self.max_children = processes


def make_server(
            socket, app=None, *, threaded=False, processes=1,
            request_handler=None, passthrough_errors=False
        ):
    """Create a new server instance listening on the given socket that is
    either threaded, or forks or just processes one request after another.
    """
    if threaded and processes > 1:
        raise TypeError(
            "cannot have a multithreaded and multi process server."
        )
    elif threaded:
        return ThreadedWSGIServer(
            socket, app, handler=request_handler,
            passthough_errors=passthrough_errors,
        )
    elif processes > 1:
        return ForkingWSGIServer(
            socket, app,  handler=request_handler,
            passthrough_errors=passthrough_errors,
            processes=processes
        )
    else:
        return BaseWSGIServer(
            socket, app, handler=request_handler,
            passthrough_errors=passthrough_errors
        )


def make_socket(
            host, port=0, family=None, type=socket.SOCK_STREAM,
            backlog=2048, blocking=True, ssl_context=None
        ):
    if host.startswith('fd://'):
        # Socket has been passed in
        if family is None:
            family = socket.AF_UNIX
        fd = int(host[len('fd://'):])
        sock = socket.fromfd(fd, family, type)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    else:
        # Create socket from scratch.
        # From a unix socket path:
        if host.startswith('unix:'):
            if family is None:
                family = socket.AF_UNIX
            if family is not socket.AF_UNIX:
                raise ValueError(
                    "got unix socket address but socket family is not AF_UNIX"
                )
            address = host[len('unix:'):]
        # From a normal ipv4 or ipv6 address:
        else:
            if family is None:
                if ':' in host:
                    family = socket.AF_INET6
                else:
                    family = socket.AF_INET
            if family is socket.AF_UNIX:
                raise ValueError("expected unix socket path")

            address = (host, port)

        sock = socket.socket(family, type)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        sock.bind(address)
        sock.listen(backlog)

    if blocking:
        sock.setblocking(1)
    else:
        sock.setblocking(0)

    if ssl_context is not None:
        if isinstance(ssl_context, tuple):
            ssl_context = load_ssl_context(*ssl_context)
        if ssl_context == 'adhoc':
            ssl_context = make_adhoc_ssl_context()
        sock = ssl_context.wrap_socket(sock, server_side=True)

    return sock


def main():
    import argparse
    import importlib
    import verktyg_server.argparse

    parser = argparse.ArgumentParser()

    verktyg_server.argparse.add_arguments(parser)

    parser.add_argument(
        'app_factory', metavar='FACTORY',
        help=(
            'module path of function to be called to generate wsgi '
            'application'
        )
    )

    args = parser.parse_args()

    module_name, factory_name = args.app_factory.split(':')
    factory = getattr(importlib.import_module(module_name), factory_name)
    application = factory()

    server = verktyg_server.argparse.make_server(args, application)
    server.serve_forever()



"""
    verktyg_server.tests
    ~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 by Ben Mather.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from verktyg_server.tests import (
    test_ssl, test_sockets, test_serving, test_testing, test_argparse,
)


loader = unittest.TestLoader()
suite = unittest.TestSuite((
    loader.loadTestsFromModule(test_ssl),
    loader.loadTestsFromModule(test_sockets),
    loader.loadTestsFromModule(test_serving),
    loader.loadTestsFromModule(test_testing),
    loader.loadTestsFromModule(test_argparse),
))



"""
    verktyg_server.tests.test_sockets
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 by Ben Mather.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest


class SocketsTestCase(unittest.TestCase):
    pass



"""
    verktyg_server.argparse
    ~~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 by Ben Mather.
    :license:
        BSD, see LICENSE for more details.
"""
import re
from collections import namedtuple
from argparse import ArgumentTypeError

import verktyg_server
import verktyg_server.ssl


_address_re = re.compile(r'''
    ^
    (?:
        (?P<scheme> [a-z]+)
        ://
    )?
    (?P<hostname>
        (?:
            [A-Za-z0-9]
            [A-Za-z0-9\-]*
        )
        (?:
            \.
            [A-Za-z0-9]
            [A-Za-z0-9\-]*
        )*
    )
    (?:
        :
        (?P<port> [0-9]+)
    )?
    $
''', re.VERBOSE)


_Address = namedtuple('Address', ['scheme', 'hostname', 'port'])


class _AddressType(object):
    def __call__(self, string):
        match = _address_re.match(string)

        if match is None:
            raise ArgumentTypeError("Invalid address %r" % string)

        scheme, hostname, port = match.group('scheme', 'hostname', 'port')

        # TODO should scheme be validated here?
        if scheme and scheme not in {'https', 'http'}:
            raise ArgumentTypeError("Invalid scheme %r" % scheme)

        if port is not None:
            port = int(port)

        return _Address(scheme, hostname, port)


def add_arguments(parser):
    """Takes an ``argparse`` parser and populates it with the arguments
    required by :func:`make_server`
    """
    group = parser.add_argument_group("Serving Options")
    addr_group = group.add_mutually_exclusive_group(required=True)
    addr_group.add_argument(
        '--socket', type=str,
        help=(
            'Path of a unix socket to listen on.  If the socket does '
            'not exist it will be created'
        )
    )
    addr_group.add_argument(
        '--address', type=_AddressType(),
        help=(
            'Hostname or address to listen on.  Can include optional port'
        )
    )
    addr_group.add_argument(
        '--fd', type=str,
        help=(
            'File descriptor to listen on'
        )
    )

    group = parser.add_argument_group("SSL Options")
    group.add_argument(
        '--certificate', type=str,
        help=(
            'Path to certificate file'
        )
    )
    group.add_argument(
        '--private-key', type=str,
        help=(
            'Path to private key file'
        )
    )


def make_server(args, application):
    """Takes an `argparse` namespace and a wsgi application and returns a
    new http server
    """
    if args.certificate:
        ssl_context = verktyg_server.ssl.load_ssl_context(
            args.certificate, args.private_key
        )
    else:
        if args.private_key:
            raise ValueError("Private key provided but no certificate")
        ssl_context = None

    if args.socket:
        raise NotImplementedError()

    elif args.address:
        scheme = args.address.scheme
        if not scheme:
            scheme = 'https' if ssl_context else 'http'

        address = args.address.hostname

        if not args.address.port:
            port = {
                'http': 80,
                'https': 443,
            }[scheme]

        if scheme == 'https' and not ssl_context:
            ssl_context = verktyg_server.ssl.make_adhoc_ssl_context()

    elif args.fd:
        scheme = 'fd'
        address = str(int(args.fd))
        port = None

    socket = verktyg_server.make_socket(
        '{scheme}://{address}'.format(scheme=scheme, address=address), port,
        ssl_context=ssl_context
    )

    server = verktyg_server.make_server(socket, application)
    return server



"""
    verktyg_server.tests.test_serving
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2015 by Ben Mather.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest

from http.client import HTTPConnection
from threading import Thread

from verktyg_server import make_socket, make_server

import logging
logging.disable(logging.CRITICAL)


class ServingTestCase(unittest.TestCase):
    def test_basic(self):
        def application(environ, start_response):
            status = '200 OK'
            headers = [('Content-type', 'text/plain; charset=utf-8')]
            start_response(status, headers)
            return [b"Hello world!"]

        socket = make_socket('localhost')
        port = socket.getsockname()[1]

        server = make_server(socket, application)
        thread = Thread(target=server.serve_forever)
        thread.start()

        try:
            conn = HTTPConnection('localhost', port)
            conn.request('GET', '/')

            resp = conn.getresponse()
            self.assertEqual(resp.read(), b"Hello world!")
        finally:
            server.shutdown()
            thread.join()

    def test_yield_no_set_headers(self):
        def application(environ, start_response):
            return [b"Who needs headers"]

        socket = make_socket('localhost')
        port = socket.getsockname()[1]

        server = make_server(socket, application)
        thread = Thread(target=server.serve_forever)
        thread.start()

        try:
            conn = HTTPConnection('localhost', port)
            conn.request('GET', '/')

            resp = conn.getresponse()
            self.assertEqual(resp.status, 500)
        finally:
            server.shutdown()
            thread.join()
